from typing import Union, Tuple, Dict
from omegaconf import DictConfig

import torch
from torch.utils.data import DataLoader

from transformers import AutoTokenizer

from data.base import BaseDataset

def make_loader(
    config: DictConfig,
    device: Union[str, int, torch.device]
) -> Tuple[DataLoader]:
    
    tok = AutoTokenizer.from_pretrained(config.model.name_or_path)
    
    train_set = zsREDataset(
        config.data,
        config.data.train_path,
        tok,
        device
    )

    valid_set = zsREDataset(
        config.data,
        config.data.valid_path,
        tok,
        device
    )
    
    train_loader = DataLoader(
        train_set,
        config.data.n_edits,
        True,
        collate_fn = train_set.collate_fn,
        drop_last = True
    )

    valid_loader = DataLoader(
        valid_set,
        config.data.n_edits,
        True,
        collate_fn = valid_set.collate_fn,
        drop_last = True
    )
    
    return train_loader, valid_loader

class zsREDataset(BaseDataset):
    
    def __getitem__(self, idx) -> Dict[str, Dict[str, torch.LongTensor]]:
        row = self.data[idx]
        
        prompt = row["src"]
        equiv_prompt = row["rephrase"]
        answer = row["ans"]
        unrel_prompt = row["loc"] + "?"
        unrel_answer = row["loc_ans"]
    
        return {
            "edit_tuples": self.tok_tuples(prompt, answer),
            "equiv_tuples": self.tok_tuples(equiv_prompt, answer),
            "unrel_tuples": self.tok_tuples(unrel_prompt, unrel_answer)
        }
        
    def tok_tuples(
        self,
        prompt: str,
        answer: str
    ) -> Dict[str, torch.LongTensor]:

        answer = " " + answer
        
        tok_prompt = self.tok(
            prompt,
            return_tensors = "pt",
        )
        tok_answer = self.tok(
            answer,
            return_tensors = "pt",
            add_special_tokens = False
        )

        tok_tuples = {
            key: torch.cat((value, tok_answer[key][:, :-1]), -1)
            for key, value in tok_prompt.items()
        }
        
        tok_tuples["labels"] = torch.cat((
            torch.full(tok_prompt["input_ids"].shape, -100)[:, 1:],
            tok_answer["input_ids"]
        ), -1)
        
        return tok_tuples